################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Plugin(Enum):
    """The plugins (grids/charts) available in Perspective.  Pass these into
    the `plugin` arg in `PerspectiveWidget` or `PerspectiveViewer`.

    Examples:
        >>> widget = PerspectiveWidget(data, plugin=Plugin.TREEMAP)
    """

    GRID = "Datagrid"

    YBAR = "Y Bar"
    XBAR = "X Bar"
    YLINE = "Y Line"
    YAREA = "Y Area"
    YSCATTER = "Y Scatter"
    XYLINE = "X/Y Line"
    XYSCATTER = "X/Y Scatter"
    TREEMAP = "Treemap"
    SUNBURST = "Sunburst"
    HEATMAP = "Heatmap"

    YBAR_D3 = "Y Bar"
    XBAR_D3 = "X Bar"
    YLINE_D3 = "Y Line"
    YAREA_D3 = "Y Area"
    YSCATTER_D3 = "Y Scatter"
    XYSCATTER_D3 = "X/Y Scatter"
    TREEMAP_D3 = "Treemap"
    SUNBURST_D3 = "Sunburst"
    HEATMAP_D3 = "Heatmap"

    CANDLESTICK = "Candlestick"
    OHLC = "OHLC"

    MAP_SCATTER = "Map Scatter"

    @staticmethod
    def options():
        return list(c.value for c in Plugin)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import asyncio
import json
from abc import ABC, abstractmethod

from .client import PerspectiveClient
from ..manager.manager_internal import DateTimeEncoder


class Periodic(ABC):
    def __init__(self, callback, interval):
        self._callback = callback
        self._interval = interval

    @abstractmethod
    async def start(self):
        ...

    @abstractmethod
    async def stop(self):
        ...


class PerspectiveWebsocketConnection(ABC):
    @abstractmethod
    async def connect(self, url, on_message, max_message_size) -> None:
        ...

    @abstractmethod
    def periodic(self, callback, interval) -> Periodic:
        ...

    @abstractmethod
    async def write(self, message, binary=False):
        ...

    @abstractmethod
    async def close(self):
        ...


class PerspectiveWebsocketClient(PerspectiveClient):
    # Ping the server every 30 seconds
    PING_TIMEOUT = 15 * 1000

    def __init__(self, websocket: PerspectiveWebsocketConnection):
        """Create a `PerspectiveWebsocketClient` that interfaces with a
        Perspective server over a Websocket"""
        super(PerspectiveWebsocketClient, self).__init__()

        self._websocket = websocket
        self._pending_binary = None
        self._pending_binary_length = 0
        self._pending_port_id = None
        self._full_binary = b""

    async def _send_ping(self):
        """Send a `ping` heartbeat message to the server's Websocket, which will
        respond with `pong`."""
        await self.send("ping")

    async def connect(self, url):
        """Connect to the remote websocket, and send the `init` message to
        assert that the Websocket is alive and accepting connections."""
        await self._websocket.connect(
            url,
            on_message=self.on_message,
            max_message_size=1024 * 1024 * 1024,
        )

        await self.send({"id": -1, "cmd": "init"})

        # Send a `ping` message every 15 seconds.
        self._ping_callback = self._websocket.periodic(
            self._send_ping,
            interval=self.PING_TIMEOUT,
        )

        await self._ping_callback.start()

    def on_message(self, msg):
        """When a message is received, send it to the `_handle` method, or
        await the incoming binary from the server."""
        if msg == "pong":
            # Do not respond to server pong heartbeats - only send them
            return

        if self._pending_binary is not None:
            binary_msg = msg

            self._full_binary += binary_msg

            if len(self._full_binary) == self._pending_binary_length:
                # Chunking is complete
                binary_msg = self._full_binary

            else:
                # Wait for the next chunk
                return

            result = {"data": {"id": self._pending_binary, "data": binary_msg}}

            if self._pending_port_id is not None:
                result["data"]["data"] = {
                    "port_id": self._pending_port_id,
                    "delta": binary_msg,
                }

            self._handle(result)

            # Clear flags for special binary message flow
            self._pending_binary = None
            self._pending_binary_length = None
            self._pending_port_id = None
            self._full_binary = b""

        elif isinstance(msg, str):
            msg = json.loads(msg)

            if msg.get("binary_length"):
                self._pending_binary = msg["id"]
                self._pending_binary_length = msg["binary_length"]

                if msg.get("data") and msg["data"].get("port_id", None) is not None:
                    self._pending_port_id = msg["data"].get("port_id")
            else:
                self._handle({"data": msg})
        else:
            # websocket client sometimes calls None on disconnect ..
            pass

    async def send(self, msg):
        """Send a message to the Websocket endpoint."""
        if isinstance(msg, dict) and msg.get("args") and len(msg["args"]) > 0 and isinstance(msg["args"][0], (bytes, bytearray)):
            msg["binary_length"] = len(msg["args"][0])
            pre_msg = msg
            binary_msg = pre_msg["args"].pop(0)

            # tornado_handler expects the first arg to be an empty object,
            # which is the result of stringifying an ArrayBuffer in JS - add
            # an empty dict here so we don't break compatibility.
            pre_msg["args"].insert(0, {})

            # Must be received in order - expectation that binary follows premsg
            await asyncio.gather(
                *[
                    self._websocket.write(json.dumps(pre_msg, cls=DateTimeEncoder)),
                    self._websocket.write(binary_msg, binary=True),
                ]
            )

        elif isinstance(msg, str):
            await self._websocket.write(msg)

        else:
            await self._websocket.write(json.dumps(msg, cls=DateTimeEncoder))

    async def terminate(self):
        """Close the websocket client connection."""
        await self._ping_callback.stop()
        await self._websocket.close()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import numpy as np
from datetime import date, datetime
from perspective.table import Table


class TestUpdate(object):
    def test_update_from_schema(self):
        tbl = Table({
            "a": str,
            "b": int
        })
        tbl.update([{"a": "abc", "b": 123}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 123}]

    def test_update_columnar_from_schema(self):
        tbl = Table({
            "a": str,
            "b": int
        })
        tbl.update({"a": ["abc"], "b": [123]})
        assert tbl.view().to_records() == [{"a": "abc", "b": 123}]

    def test_update_csv(self):
        tbl = Table({
            "a": str,
            "b": int
        })

        view = tbl.view()
        tbl.update("a,b\nxyz,123\ndef,100000000")

        assert view.to_dict() == {
            "a": ["xyz", "def"],
            "b": [123, 100000000]
        }

    def test_update_csv_indexed(self):
        tbl = Table({
            "a": str,
            "b": float
        }, index="a")

        view = tbl.view()
        tbl.update("a,b\nxyz,1.23456718\ndef,100000000.1")

        assert view.to_dict() == {
            "a": ["def", "xyz"],
            "b": [100000000.1, 1.23456718]
        }

        tbl.update("a,b\nxyz,0.00000001\ndef,1234.5678\nefg,100.2")

        assert view.to_dict() == {
            "a": ["def", "efg", "xyz"],
            "b": [1234.5678, 100.2, 0.00000001]
        }



    def test_update_append(self):
        tbl = Table([{"a": "abc", "b": 123}])
        tbl.update([{"a": "def", "b": 456}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 123}, {"a": "def", "b": 456}]

    def test_update_partial(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        tbl.update([{"a": "abc", "b": 456}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 456}]

    def test_update_partial_add_row(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        tbl.update([{"a": "abc", "b": 456}, {"a": "def"}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 456}, {"a": "def", "b": None}]

    def test_update_partial_noop(self):
        tbl = Table([{"a": "abc", "b": 1, "c": 2}], index="a")
        tbl.update([{"a": "abc", "b": 456}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 456, "c": 2}]

    def test_update_partial_unset(self):
        tbl = Table([{"a": "abc", "b": 1, "c": 2}, {"a": "def", "b": 3, "c": 4}], index="a")
        tbl.update([{"a": "abc"}, {"a": "def", "c": None}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 1, "c": 2}, {"a": "def", "b": 3, "c": None}]

    def test_update_columnar_partial_add_row(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")

        tbl.update({
            "a": ["abc", "def"],
            "b": [456, None]
        })

        assert tbl.view().to_records() == [{"a": "abc", "b": 456}, {"a": "def", "b": None}]

    def test_update_columnar_partial_noop(self):
        tbl = Table([{"a": "abc", "b": 1, "c": 2}], index="a")

        # no-op because "c" is not in the update dataset
        tbl.update({
            "a": ["abc"],
            "b": [456]
        })

        assert tbl.view().to_records() == [{"a": "abc", "b": 456, "c": 2}]

    def test_update_columnar_partial_unset(self):
        tbl = Table([{"a": "abc", "b": 1, "c": 2}, {"a": "def", "b": 3, "c": 4}], index="a")

        tbl.update({
            "a": ["abc"],
            "b": [None]
        })

        assert tbl.view().to_records() == [{"a": "abc", "b": None, "c": 2}, {"a": "def", "b": 3, "c": 4}]

    def test_update_partial_subcolumn(self):
        tbl = Table([{"a": "abc", "b": 123, "c": 456}], index="a")
        tbl.update([{"a": "abc", "c": 1234}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 123, "c": 1234}]

    def test_update_partial_subcolumn_dict(self):
        tbl = Table([{"a": "abc", "b": 123, "c": 456}], index="a")
        tbl.update({
            "a": ["abc"],
            "c": [1234]
        })
        assert tbl.view().to_records() == [{"a": "abc", "b": 123, "c": 1234}]

    def test_update_partial_cols_more_columns_than_table(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        tbl.update([{"a": "abc", "b": 456, "c": 789}])
        assert tbl.view().to_records() == [{"a": "abc", "b": 456}]

    def test_update_columnar_append(self):
        tbl = Table({"a": ["abc"], "b": [123]})
        tbl.update({"a": ["def"], "b": [456]})
        assert tbl.view().to_records() == [{"a": "abc", "b": 123}, {"a": "def", "b": 456}]

    def test_update_columnar_partial(self):
        tbl = Table({"a": ["abc"], "b": [123]}, index="a")
        tbl.update({"a": ["abc"], "b": [456]})
        assert tbl.view().to_records() == [{"a": "abc", "b": 456}]

    # make sure already created views are notified properly
    def test_update_from_schema_notify(self):
        tbl = Table({
            "a": str,
            "b": int
        })
        view = tbl.view()
        assert view.num_rows() == 0
        tbl.update([{"a": "abc", "b": 123}])
        assert view.to_records() == [{"a": "abc", "b": 123}]

    def test_update_columnar_from_schema_notify(self):
        tbl = Table({
            "a": str,
            "b": int
        })
        view = tbl.view()
        assert view.num_rows() == 0
        tbl.update({"a": ["abc"], "b": [123]})
        assert view.to_records() == [{"a": "abc", "b": 123}]

    def test_update_append_notify(self):
        tbl = Table([{"a": "abc", "b": 123}])
        view = tbl.view()
        assert view.num_rows() == 1
        tbl.update([{"a": "def", "b": 456}])
        assert view.to_records() == [{"a": "abc", "b": 123}, {"a": "def", "b": 456}]

    def test_update_partial_notify(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        view = tbl.view()
        assert view.num_rows() == 1
        tbl.update([{"a": "abc", "b": 456}])
        assert view.to_records() == [{"a": "abc", "b": 456}]

    def test_update_partial_cols_not_in_schema_notify(self):
        tbl = Table([{"a": "abc", "b": 123}], index="a")
        view = tbl.view()
        assert view.num_rows() == 1
        tbl.update([{"a": "abc", "b": 456, "c": 789}])
        assert view.to_records() == [{"a": "abc", "b": 456}]

    def test_update_columnar_append_notify(self):
        tbl = Table({"a": ["abc"], "b": [123]})
        view = tbl.view()
        assert view.num_rows() == 1
        tbl.update({"a": ["def"], "b": [456]})
        assert view.to_records() == [{"a": "abc", "b": 123}, {"a": "def", "b": 456}]

    def test_update_columnar_partial_notify(self):
        tbl = Table({"a": ["abc"], "b": [123]}, index="a")
        view = tbl.view()
        assert view.num_rows() == 1
        tbl.update({"a": ["abc"], "b": [456]})
        assert view.to_records() == [{"a": "abc", "b": 456}]

    # bool

    def test_update_bool_from_schema(self):
        bool_data = [{"a": True, "b": False}, {"a": True, "b": True}]
        tbl = Table({
            "a": bool,
            "b": bool
        })
        tbl.update(bool_data)
        assert tbl.size() == 2
        assert tbl.view().to_records() == bool_data

    def test_update_bool_str_from_schema(self):
        bool_data = [{"a": "True", "b": "False"}, {"a": "True", "b": "True"}]
        tbl = Table({
            "a": bool,
            "b": bool
        })
        tbl.update(bool_data)
        assert tbl.size() == 2
        assert tbl.view().to_records() == [
            {"a": True, "b": False},
            {"a": True, "b": True}]

    def test_update_bool_str_all_formats_from_schema(self):
        bool_data = [
            {"a": "True", "b": "False"},
            {"a": "t", "b": "f"},
            {"a": "true", "b": "false"},
            {"a": 1, "b": 0},
            {"a": "on", "b": "off"}
        ]
        tbl = Table({
            "a": bool,
            "b": bool
        })
        tbl.update(bool_data)
        assert tbl.size() == 5
        assert tbl.view().to_dict() == {
            "a": [True, True, True, True, True],
            "b": [False, False, False, False, False]
        }

    def test_update_bool_int_from_schema(self):
        bool_data = [{"a": 1, "b": 0}, {"a": 1, "b": 0}]
        tbl = Table({
            "a": bool,
            "b": bool
        })
        tbl.update(bool_data)
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "a": [True, True],
            "b": [False, False]
        }

    # dates and datetimes
    def test_update_date(self):
        tbl = Table({"a": [date(2019, 7, 11)]})
        tbl.update([{"a": date(2019, 7, 12)}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11)},
            {"a": datetime(2019, 7, 12)}
        ]

    def test_update_date_np(self):
        tbl = Table({"a": [date(2019, 7, 11)]})
        tbl.update([{"a": np.datetime64(date(2019, 7, 12))}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11)},
            {"a": datetime(2019, 7, 12)}
        ]

    def test_update_datetime(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": datetime(2019, 7, 12, 11, 0)}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11, 11, 0)},
            {"a": datetime(2019, 7, 12, 11, 0)}
        ]

    def test_update_datetime_np(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": np.datetime64(datetime(2019, 7, 12, 11, 0))}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11, 11, 0)},
            {"a": datetime(2019, 7, 12, 11, 0)}
        ]

    def test_update_datetime_np_ts(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": np.datetime64("2019-07-12T11:00")}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11, 11, 0)},
            {"a": datetime(2019, 7, 12, 11, 0)}
        ]

    def test_update_datetime_timestamp_seconds(self, util):
        ts = util.to_timestamp(datetime(2019, 7, 12, 11, 0, 0))
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": ts}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11, 11, 0)},
            {"a": datetime(2019, 7, 12, 11, 0)}
        ]

    def test_update_datetime_timestamp_ms(self, util):
        ts = util.to_timestamp(datetime(2019, 7, 12, 11, 0, 0))
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": ts}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 11, 11, 0)},
            {"a": datetime(2019, 7, 12, 11, 0)}
        ]

    # partial date & datetime updates

    def test_update_date_partial(self):
        tbl = Table({"a": [date(2019, 7, 11)], "b": [1]}, index="b")
        tbl.update([{"a": date(2019, 7, 12), "b": 1}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12), "b": 1}]

    def test_update_date_np_partial(self):
        tbl = Table({"a": [date(2019, 7, 11)], "b": [1]}, index="b")
        tbl.update([{"a": np.datetime64(date(2019, 7, 12)), "b": 1}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12), "b": 1}]

    def test_update_datetime_partial(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)], "b": [1]}, index="b")
        tbl.update([{"a": datetime(2019, 7, 12, 11, 0), "b": 1}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0), "b": 1}]

    def test_update_datetime_np_partial(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)], "b": [1]}, index="b")
        tbl.update([{"a": np.datetime64(datetime(2019, 7, 12, 11, 0)), "b": 1}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0), "b": 1}]

    def test_update_datetime_np_ts_partial(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)], "b": [1]}, index="b")
        tbl.update([{"a": np.datetime64("2019-07-12T11:00"), "b": 1}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0), "b": 1}]

    def test_update_datetime_timestamp_seconds_partial(self, util):
        ts = util.to_timestamp(datetime(2019, 7, 12, 11, 0, 0))
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)], "idx": [1]}, index="idx")
        tbl.update([{"a": ts, "idx": 1}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 12, 11, 0), "idx": 1}
        ]

    def test_update_datetime_timestamp_ms_partial(self, util):
        ts = util.to_timestamp(datetime(2019, 7, 12, 11, 0, 0))
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)], "idx": [1]}, index="idx")
        tbl.update([{"a": ts, "idx": 1}])
        assert tbl.view().to_records() == [
            {"a": datetime(2019, 7, 12, 11, 0), "idx": 1}
        ]

    # updating dates using implicit index

    def test_update_date_partial_implicit(self):
        tbl = Table({"a": [date(2019, 7, 11)]})
        tbl.update([{"a": date(2019, 7, 12), "__INDEX__": 0}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12)}]

    def test_update_date_np_partial_implicit(self):
        tbl = Table({"a": [date(2019, 7, 11)]})
        tbl.update([{"a": np.datetime64(date(2019, 7, 12)), "__INDEX__": 0}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12)}]

    def test_update_datetime_partial_implicit(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": datetime(2019, 7, 12, 11, 0), "__INDEX__": 0}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0)}]

    def test_update_datetime_np_partial_implicit(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": np.datetime64(datetime(2019, 7, 12, 11, 0)), "__INDEX__": 0}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0)}]

    def test_update_datetime_np_ts_partial_implicit(self):
        tbl = Table({"a": [datetime(2019, 7, 11, 11, 0)]})
        tbl.update([{"a": np.datetime64("2019-07-12T11:00"), "__INDEX__": 0}])
        assert tbl.view().to_records() == [{"a": datetime(2019, 7, 12, 11, 0)}]

    # implicit index

    def test_update_implicit_index(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data)
        view = tbl.view()
        tbl.update([{
            "__INDEX__": [0],
            "a": 3,
            "b": 15
        }])
        assert view.to_records() == [{"a": 3, "b": 15}, {"a": 2, "b": 3}]

    def test_update_implicit_index_dict_noop(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data)
        view = tbl.view()

        # "b" does not exist in dataset, so no-op
        tbl.update({
            "__INDEX__": [0],
            "a": [3],
        })
        assert view.to_records() == [{"a": 3, "b": 2}, {"a": 2, "b": 3}]

    def test_update_implicit_index_dict_unset_with_null(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data)
        view = tbl.view()
        
        # unset because "b" is null
        tbl.update({
            "__INDEX__": [0],
            "a": [3],
            "b": [None]
        })
        assert view.to_records() == [{"a": 3, "b": None}, {"a": 2, "b": 3}]

    def test_update_implicit_index_multi(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}, {"a": 4, "b": 5}]
        tbl = Table(data)
        view = tbl.view()
        tbl.update([{
            "__INDEX__": [0],
            "a": 3,
        }, {
            "__INDEX__": [2],
            "a": 5
        }])
        assert view.to_records() == [{"a": 3, "b": 2}, {"a": 2, "b": 3}, {"a": 5, "b": 5}]

    def test_update_implicit_index_symmetric(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data)
        view = tbl.view()
        records = view.to_records(index=True)
        idx = records[0]["__INDEX__"]
        tbl.update([{
            "__INDEX__": idx,
            "a": 3
        }])
        assert view.to_records() == [{"a": 3, "b": 2}, {"a": 2, "b": 3}]

    def test_update_explicit_index(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "a": 1,
            "b": 3
        }])
        assert view.to_records() == [{"a": 1, "b": 3}, {"a": 2, "b": 3}]

    def test_update_explicit_index_multi(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}, {"a": 3, "b": 4}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "a": 1,
            "b": 3
        }, {
            "a": 3,
            "b": 5
        }])
        assert view.to_records() == [{"a": 1, "b": 3}, {"a": 2, "b": 3}, {"a": 3, "b": 5}]

    def test_update_explicit_index_multi_append(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}, {"a": 3, "b": 4}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "a": 1,
            "b": 3
        }, {
            "a": 12,
            "b": 5
        }])
        assert view.to_records() == [{"a": 1, "b": 3}, {"a": 2, "b": 3}, {"a": 3, "b": 4}, {"a": 12, "b": 5}]

    def test_update_explicit_index_multi_append_noindex(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}, {"a": 3, "b": 4}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "a": 1,
            "b": 3
        }, {
            "b": 5
        }])
        assert view.to_records() == [{"a": None, "b": 5}, {"a": 1, "b": 3}, {"a": 2, "b": 3}, {"a": 3, "b": 4}]

    def test_update_implicit_index_with_explicit_unset(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "__INDEX__": [1],
            "b": 3
        }])
        assert view.to_records() == [{"a": 1, "b": 3}, {"a": 2, "b": 3}]

    def test_update_implicit_index_with_explicit_set(self):
        data = [{"a": 1, "b": 2}, {"a": 2, "b": 3}]
        tbl = Table(data, index="a")
        view = tbl.view()
        tbl.update([{
            "__INDEX__": [1],
            "a": 1,  # should ignore re-specification of pkey
            "b": 3
        }])
        assert view.to_records() == [{"a": 1, "b": 3}, {"a": 2, "b": 3}]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Aggregate(Enum):
    """The aggregation operators available in Perspective. Pass these into the
    `aggregates` arg in `PerspectiveWidget` or `PerspectiveViewer`.

    Examples:
        >>> widget = PerspectiveWidget(data, aggregates={"a": Aggregate.LAST})
    """

    AND = "and"
    ANY = "any"
    AVG = "avg"
    COUNT = "count"
    DISTINCT_COUNT = "distinct count"
    DISTINCT_LEAF = "distinct leaf"
    DOMINANT = "dominant"
    FIRST_BY_INDEX = "first by index"
    LAST_BY_INDEX = "last by index"
    LAST_MINUS_FIRST = "last minus first"
    LAST = "last"
    HIGH = "high"
    JOIN = "join"
    LOW = "low"
    HIGH_MINUS_LOW = "high minus low"
    MEAN = "mean"
    MEDIAN = "median"
    OR = "or"
    PCT_SUM_PARENT = "pct sum parent"
    PCT_SUM_GRAND_TOTAL = "pct sum grand total"
    STANDARD_DEVIATION = "stddev"
    SUM = "sum"
    SUM_ABS = "sum abs"
    ABS_SUM = "abs sum"
    SUM_NOT_NULL = "sum not null"
    UNIQUE = "unique"
    VARIANCE = "var"

    @staticmethod
    def options():
        return list(c.value for c in Aggregate)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from random import random

from ..libpsp import is_libpsp
from .validate import (
    validate_aggregates,
    validate_split_by,
    validate_columns,
    validate_expressions,
    validate_filter,
    validate_plugin,
    validate_plugin_config,
    validate_group_by,
    validate_sort,
)
from .viewer_traitlets import PerspectiveTraitlets

if is_libpsp():
    from ..core.exception import PerspectiveError
    from ..libpsp import PerspectiveManager, Table, View


class PerspectiveViewer(PerspectiveTraitlets, object):
    """PerspectiveViewer wraps the `perspective.Table` API and exposes an API
    around creating views, loading data, and updating data.
    """

    # Keep track of attributes that can be set via Enum, and their
    # validation methods.
    ENUM_VALIDATORS = {
        "plugin": validate_plugin,
        "aggregates": validate_aggregates,
        "sort": validate_sort,
    }

    # Viewer attributes that should be saved in `save()` and restored using
    # `restore()`. Symmetric to `PERSISTENT_ATTRIBUTES` in `perspective-viewer`.
    PERSISTENT_ATTRIBUTES = (
        "group_by",
        "split_by",
        "filter",
        "sort",
        "aggregates",
        "columns",
        "expressions",
        "plugin",
        "plugin_config",
        "theme",
        "settings",
        "title",
    )

    def __init__(
        self,
        plugin="Datagrid",
        columns=None,
        group_by=None,
        split_by=None,
        aggregates=None,
        sort=None,
        filter=None,
        expressions=None,
        plugin_config=None,
        settings=True,
        theme=None,
        title=None,
    ):
        """Initialize an instance of `PerspectiveViewer` with the given viewer
        configuration.  Do not pass a `Table` or data into the constructor -
        use the :func:`load()` method to provide the viewer with data.

        Keyword Arguments:
            columns (:obj:`list` of :obj:`str`): A list of column names to be
                visible to the user.
            group_by (:obj:`list` of :obj:`str`): A list of column names to
                use as group by.
            split_by (:obj:`list` of :obj:`str`): A list of column names
                to use as split by.
            aggregates (:obj:`dict` of :obj:`str` to :obj:`str`):  A dictionary
                of column names to aggregate types, which specify aggregates
                for individual columns.
            sort (:obj:`list` of :obj:`list` of :obj:`str`): A list of lists,
                each list containing a column name and a sort direction
                (``asc``, ``desc``, ``asc abs``, ``desc abs``, ``col asc``,
                ``col desc``, ``col asc abs``, ``col desc abs``).
            filter (:obj:`list` of :obj:`list` of :obj:`str`): A list of lists,
                each list containing a column name, a filter comparator, and a
                value to filter by.
            expressions (:obj:`list` of :obj:`str`): A list of string
                expressions which are applied to the view.
            plugin (:obj:`str`/:obj:`perspective.Plugin`): Which plugin to
                select by default.
            plugin_config (:obj:`dict`): A configuration for the plugin, i.e.
                the datagrid plugin or a chart plugin.
            settings(:obj:`bool`): Whether the perspective query settings
                panel should be open.
            theme (:obj:`str`): The color theme to use.

        Examples:
            >>> viewer = PerspectiveViewer(
            ...     aggregates={"a": "avg"},
            ...     group_by=["a"],
            ...     sort=[["b", "desc"]],
            ...     filter=[["a", ">", 1]],
            ...     expressions=["// new column \n \"Sales\" + \"Profit\""]
            ... )
        """

        # Create an instance of `PerspectiveManager`, which receives messages
        # from the `PerspectiveJupyterClient` on the front-end.
        if is_libpsp():
            self.manager = PerspectiveManager()

        # The string name of the Table under management by this viewer and its
        # attached PerspectiveManager
        self.table_name = None

        # Viewer configuration
        self.plugin = validate_plugin(plugin)
        self.columns = validate_columns(columns) or []
        self.group_by = validate_group_by(group_by) or []
        self.split_by = validate_split_by(split_by) or []
        self.aggregates = validate_aggregates(aggregates) or {}
        self.sort = validate_sort(sort) or []
        self.filter = validate_filter(filter) or []
        self.expressions = validate_expressions(expressions) or []
        self.plugin_config = validate_plugin_config(plugin_config) or {}
        self.settings = settings
        self.theme = theme
        self.title = title

    @property
    def table(self):
        """Returns the ``perspective.Table`` under management by the viewer."""
        return self.manager.get_table(self.table_name)

    def load(self, data, **options):
        """Given a ``perspective.Table``, a ``perspective.View``,
        or data that can be handled by ``perspective.Table``, pass it to the
        viewer.  Like `__init__`, load accepts a `perspective.Table`, a dataset,
        or a schema. If running in client mode, `load` defers to the browser's
        Perspective engine. This means that loading Python-only datasets,
        especially ones that cannot be serialized into JSON, may cause some
        issues.

        ``load()`` resets the state of the viewer :

        * If a ``perspective.Table`` has already been  loaded, ``**options`` is
          ignored as the options already set on the ``Table`` take precedence.

        * If a ``perspective.View`` is loaded, the options on the
          ``perspective.Table`` linked to the view take precedence.

        If data is passed in, a ``perspective.Table`` is automatically created
        by this method, and the options passed to ``**config`` are extended to
        the new Table.  If the widget already has a dataset, and the new data
        has different columns to the old one, then the widget state (pivots,
        sort, etc.) is cleared to prevent applying settings on columns that
        don't exist.

        Args:
            data (:obj:`Table`|:obj:`View`|:obj:`dict`|:obj:`list`|:obj:`pandas.DataFrame`|:obj:`bytes`|:obj:`str`): a
                `perspective.Table` instance, a `perspective.View` instance, or
                a dataset to be loaded in the viewer.

        Keyword Arguments:
            name (:obj:`str`): An optional name to reference the table by so it can
                be accessed from the front-end. If not provided, a name will
                be generated.
            index (:obj:`str`): A column name to be used as the primary key.
                Ignored if a ``Table`` or ``View`` is supplied.
            limit (:obj:`int`): A upper limit on the number of rows in the Table.
                Cannot be set at the same time as `index`. Ignored if a
                ``Table`` or ``View`` is supplied.

        Examples:
            >>> from perspective import Table, PerspectiveViewer
            >>> data = {"a": [1, 2, 3]}
            >>> tbl = Table(data)
            >>> viewer = PerspectiveViewer()
            >>> viewer.load(tbl)
            >>> viewer.load(data, index="a") # viewer state is reset
            >>> viewer2 = PerspectiveViewer()
            >>> viewer2.load(tbl.view())

        """
        name = options.pop("name", str(random()))

        if isinstance(data, Table):
            table = data
        elif isinstance(data, View):
            raise PerspectiveError("Only `Table` or data can be loaded.")
        else:
            table = Table(data, **options)

        self.manager.host_table(name, table)

        # Reset the viewer when `load()` is called multiple times.
        if self.table_name is not None:
            self.reset()

        # If the user does not set columns to show, synchronize viewer state
        # with dataset.
        if len(self.columns) == 0:
            self.columns = table.columns()

        self.table_name = name

    def update(self, data):
        """Update the table under management by the viewer with new data.
        This function follows the semantics of `Table.update()`, and will be
        affected by whether an index is set on the underlying table.

        Args:
            data (:obj:`dict`|:obj:`list`|:obj:`pandas.DataFrame`): the
            update data for the table.
        """
        self.table.update(data)

    def clear(self):
        """Clears the rows of this viewer's ``Table``."""
        if self.table is not None:
            self.table.clear()

    def replace(self, data):
        """Replaces the rows of this viewer's `Table` with new data.

        Args:
            data (:obj:`dict`|:obj:`list`|:obj:`pandas.DataFrame`): new data
            to set into the table - must conform to the table's schema.
        """
        if self.table is not None:
            self.table.replace(data)

    def save(self):
        """Get the viewer's attribute as a dictionary, symmetric with `restore`
        so that a viewer's configuration can be reproduced."""
        return {attr: getattr(self, attr) for attr in PerspectiveViewer.PERSISTENT_ATTRIBUTES}

    def restore(self, **kwargs):
        """Restore a given set of attributes, passed as kwargs
        (e.g. dictionary). Symmetric with `save` so that a given viewer's
        configuration can be reproduced."""
        for k, v in kwargs.items():
            if k in PerspectiveViewer.PERSISTENT_ATTRIBUTES:
                setattr(self, k, v)

    def reset(self):
        """Resets the viewer's attributes and state, but does not delete or
        modify the underlying `Table`.

        Example:
            widget = PerspectiveWidget(data, group_by=["date"], plugin=Plugin.XBAR)
            widget.reset()
            widget.plugin  #
        """
        self.group_by = []
        self.split_by = []
        self.filter = []
        self.sort = []
        self.expressions = []
        self.aggregates = {}
        self.columns = []
        self.plugin = "Datagrid"
        self.plugin_config = {}

    def delete(self, delete_table=True):
        """Delete the Viewer's data and clears its internal state. If
        ``delete_table`` is True, the underlying `perspective.Table` and the
        internal `View` object will be deleted.

        Args:
            delete_table (:obj:`bool`) : whether the underlying `Table` will be
                deleted. Defaults to True.
        """
        if delete_table:
            # Delete all created views on the widget's manager instance
            for view in self.manager._views.values():
                view.delete()

            # Reset view cache
            self.manager._views = {}

            # Delete table
            self.table.delete()
            self.manager._tables.pop(self.table_name)
            self.table_name = None

        self.reset()

    def __setattr__(self, name, value):
        """Override __setattr__ in order to allow Enums to be validated
        through Traitlets."""
        if name in PerspectiveViewer.ENUM_VALIDATORS:
            # call the validator and set
            validated = PerspectiveViewer.ENUM_VALIDATORS[name](value)
            value = validated
        super(PerspectiveViewer, self).__setattr__(name, value)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import asyncio
from abc import ABC, abstractmethod
import functools

from ..core.exception import PerspectiveError


class PerspectiveHandlerBase(ABC):
    def __init__(self, **kwargs):
        """Create a new instance of the PerspectiveHandlerBase with the
        given Manager instance.

        Keyword Args:
            manager (:obj:`PerspectiveManager`): A `PerspectiveManager` instance.
                Must be provided on initialization.
            check_origin (:obj`bool`): If True, all requests will be accepted
                regardless of origin. Defaults to False.
            chunk_size (:obj:`int`): Binary messages will not exceed this length
                (in bytes);  payloads above this limit will be chunked
                across multiple messages. Defaults to `16_777_216` (16MB), and
                be disabled entirely with `None`.
        """
        self._manager = kwargs.pop("manager", None)
        if self._manager is None:
            raise PerspectiveError("A `PerspectiveManager` instance must be provided to the handler!")

        self._check_origin = kwargs.pop("check_origin", False)

        # uvicorn (used by aiohttp / starlette) defaults to 16MB max payload
        # https://github.com/encode/uvicorn/pull/995/files
        # https://github.com/euri10/uvicorn/blob/8a9369439c15a657a7726eb291858f72d3199c60/tests/protocols/test_websocket.py#L463
        self._chunk_size = kwargs.pop("chunk_size", 16_777_216)

        # https://www.tornadoweb.org/en/stable/gen.html#tornado.gen.moment
        self._chunk_sleep = kwargs.pop("_chunk_sleep", 0)

        self._session = self._manager.new_session()
        self._stream_lock = asyncio.Lock()

    def check_origin(self, origin):
        """Returns whether the handler allows requests from origins outside
        of the host URL.

        Args:
            origin (:obj"`bool`): a boolean that indicates whether requests
                outside of the host URL should be accepted. If :obj:`True`, request
                URLs will not be validated and all requests will be allowed.
                Defaults to :obj:`False`.
        """
        return self._check_origin

    def _session_callback(self, loop, *args, **kwargs):
        return asyncio.run_coroutine_threadsafe(self.post(*args, **kwargs), loop)

    async def on_message(self, message):
        """When the websocket receives a message, send it to the :obj:`process`
        method of the `PerspectiveManager` with a reference to the :obj:`post`
        callback.
        """
        if message == "ping":
            # Respond to ping heartbeats from the Websocket client.
            await self.post("pong")
            return

        loop = asyncio.get_event_loop()
        self._session.process(message, functools.partial(self._session_callback, loop))

    async def post(self, message, binary=False):
        """When `post` is called by `PerspectiveManager`, serialize the data to
        JSON and send it to the client.

        Args:
            message (:obj:`str`): a JSON-serialized string containing a message to the
                front-end `perspective-viewer`.
        """
        # Only send message in chunks if it passes the threshold set by the
        # `PerspectiveManager`.
        chunked = len(message) > self._chunk_size

        async with self._stream_lock:
            if binary and chunked:
                start = 0

                while start < len(message):
                    end = start + self._chunk_size
                    if end >= len(message):
                        end = len(message)

                    asyncio.ensure_future(self.write_message(message[start:end], binary=True))
                    start = end

                    # Allow the loop to process heartbeats so that client sockets don't
                    # get closed in the middle of sending a chunk.
                    await asyncio.sleep(self._chunk_sleep)
            else:
                await self.write_message(message, binary)

    def on_close(self):
        """Remove the views associated with the client when the websocket
        closes.
        """
        self._session.close()

    @abstractmethod
    async def write_message(self, message: str, binary: bool = False) -> None:
        """Websocket-specific implementation of writing a message to the websocket client.
        Must support writing either text or binary messages. Should only be called by this
        class's `post` function, which handles all the perspective-specific logic

        Args:
            message (str): the message to write
            binary (bool, optional): whether or not to write as binary buffer
        """



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import queue
import random
import threading
from functools import partial

import tornado.ioloop
from perspective import PerspectiveError, PerspectiveManager, Table
from pytest import raises


def syncify(f):
    """Given a function `f` that must be run on `TestAsync.loop`, queue `f` on
    the loop, block until it is evaluated, and return the result.
    """
    sem = queue.Queue()

    def _syncify_task():
        assert threading.current_thread().ident == TestAsync.thread.ident
        result = f()
        TestAsync.loop.add_callback(lambda: sem.put(result))

    def _syncify():
        TestAsync.loop.add_callback(_syncify_task)
        return sem.get()

    return _syncify


data = [{"a": i, "b": i * 0.5, "c": str(i)} for i in range(10)]


class TestAsync(object):
    @classmethod
    def setup_class(cls):
        cls.loop = tornado.ioloop.IOLoop()
        cls.loop.make_current()
        cls.thread = threading.Thread(target=cls.loop.start)
        cls.thread.daemon = True
        cls.thread.start()

    @classmethod
    def teardown_class(cls):
        cls.loop.add_callback(lambda: tornado.ioloop.IOLoop.current().stop())
        cls.loop.clear_current()
        cls.thread.join()
        cls.loop.close(all_fds=True)

    @classmethod
    def loop_is_running(cls):
        return cls.loop.asyncio_loop.is_running()

    def test_async_queue_process(self):
        tbl = Table({"a": int, "b": float, "c": str})
        manager = PerspectiveManager()
        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager.host(tbl)

        @syncify
        def _task():
            assert tbl.size() == 0
            for i in range(5):
                tbl.update([data[i]])
            return tbl.size()

        assert _task() == 5
        tbl.delete()

    def test_async_queue_process_csv(self):
        """Make sure GIL release during CSV loading works"""     
        tbl = Table("x,y,z\n1,a,true\n2,b,false\n3,c,true\n4,d,false")
        manager = PerspectiveManager()
        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager.host(tbl)

        @syncify
        def _task():
            assert tbl.size() == 4
            for i in range(5):
                tbl.update("x,y,z\n1,a,true\n2,b,false\n3,c,true\n4,d,false")
            return tbl.size()

        assert _task() == 24

        tbl.delete()

    def test_async_call_loop(self):
        tbl = Table({"a": int, "b": float, "c": str})
        manager = PerspectiveManager()
        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager.call_loop(tbl.update, data)
        manager.host(tbl)

        @syncify
        def _task():
            return tbl.size()

        assert _task() == 10
        tbl.delete()

    def test_async_call_loop_error_if_no_loop(self):
        tbl = Table({"a": int, "b": float, "c": str})
        manager = PerspectiveManager()

        with raises(PerspectiveError):
            # loop not set - errors
            manager.call_loop(tbl.update, data)

        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager.call_loop(tbl.update, data)
        manager.host(tbl)

        @syncify
        def _task():
            return tbl.size()

        # subsequent calls to call_loop will work if loop_callback is set.
        assert _task() == 10

        tbl.delete()

    def test_async_multiple_managers_queue_process(self):
        tbl = Table({"a": int, "b": float, "c": str})
        tbl2 = Table({"a": int, "b": float, "c": str})

        manager = PerspectiveManager()
        manager2 = PerspectiveManager()

        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager2.set_loop_callback(TestAsync.loop.add_callback)

        @syncify
        def _update_task():
            for i in range(5):
                tbl.update([data[i]])
                tbl2.update([data[i]])
            return tbl.size()

        assert _update_task() == 5

        @syncify
        def _flush_to_process():
            view = tbl2.view()
            records = view.to_records()
            for i in range(5):
                tbl2.update([data[i]])

            view.delete()
            return records

        assert _flush_to_process() == data[:5]

        @syncify
        def _delete_task():
            tbl2.delete()
            tbl.delete()

        _delete_task()

    def test_async_multiple_managers_mixed_queue_process(self):
        sentinel = {"called": 0}

        def sync_queue_process(f, *args, **kwargs):
            sentinel["called"] += 1
            f(*args, **kwargs)

        tbl = Table({"a": int, "b": float, "c": str})
        tbl2 = Table({"a": int, "b": float, "c": str})
        manager = PerspectiveManager()
        manager2 = PerspectiveManager()
        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        # manager uses tornado, manager2 is synchronous
        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager2.set_loop_callback(sync_queue_process)

        @syncify
        def _tbl_task():
            for i in range(5):
                tbl.update([data[i]])
            return tbl.size()

        assert _tbl_task() == 5

        for i in range(5):
            tbl2.update([data[i]])

        assert sentinel["called"] == 6

        @syncify
        def _tbl_task2():
            view = tbl.view()
            records = view.to_records()
            view.delete()
            tbl.delete()
            return records

        assert _tbl_task2() == data[:5]

        view = tbl2.view()
        assert view.to_records() == data[:5]

        view.delete()
        tbl2.delete()

    def test_async_multiple_managers_delayed_process(self):
        sentinel = {"async": 0, "sync": 0}

        def _counter(key, f, *args, **kwargs):
            sentinel[key] += 1
            return f(*args, **kwargs)

        short_delay_queue_process = partial(_counter, "sync")
        long_delay_queue_process = partial(
            TestAsync.loop.add_timeout, 1, _counter, "async"
        )

        tbl = Table({"a": int, "b": float, "c": str})
        tbl2 = Table({"a": int, "b": float, "c": str})

        manager = PerspectiveManager()
        manager2 = PerspectiveManager()
        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        manager.set_loop_callback(short_delay_queue_process)
        manager2.set_loop_callback(long_delay_queue_process)

        @syncify
        def _tbl_task():
            for i in range(10):
                tbl2.update([data[i]])

        _tbl_task()
        for i in range(10):
            tbl.update([data[i]])

        @syncify
        def _tbl_task2():
            size = tbl2.size()
            tbl2.delete()
            return size

        assert _tbl_task2() == 10
        assert tbl.size() == 10
        assert sentinel["async"] == 2
        assert sentinel["sync"] == 11

        tbl.delete()

    def test_async_single_manager_tables_chained(self):
        columns = {"index": int, "num1": int, "num2": int}
        manager = PerspectiveManager()
        tbl = Table(columns, index="index")
        view = tbl.view()
        tbl2 = Table(view.to_arrow(), index=tbl.get_index())
        manager.host(tbl, "tbl")
        manager.host(tbl2, "tbl2")
        view.on_update(lambda port, delta: tbl2.update(delta), "row")
        manager.set_loop_callback(TestAsync.loop.add_callback)

        for i in range(1000):
            manager.call_loop(tbl.update, [{"index": i, "num1": i, "num2": 2 * i}])
            i += 1

        q = queue.Queue()
        manager.call_loop(q.put, True)
        q.get()

        @syncify
        def _tbl_task2():
            size = tbl2.size()
            return size

        assert _tbl_task2() == 1000
        view.delete()
        tbl.delete()
        tbl2.delete()

    def test_async_queue_process_multiple_ports(self):
        tbl = Table({"a": int, "b": float, "c": str})
        port_ids = [0]
        port_data = [{"a": 0, "b": 0, "c": "0"}]

        for i in range(10):
            port_id = tbl.make_port()
            port_ids.append(port_id)
            port_data.append({"a": port_id, "b": port_id * 1.5, "c": str(port_id)})

        assert port_ids == list(range(0, 11))
        manager = PerspectiveManager()
        manager.host(tbl)
        manager.set_loop_callback(TestAsync.loop.add_callback)

        assert syncify(lambda: tbl.size())() == 0

        random.shuffle(port_ids)

        @syncify
        def _tbl_task():
            for port_id in port_ids:
                idx = port_id if port_id < len(port_ids) else len(port_ids) - 1
                tbl.update([port_data[idx]], port_id=port_id)
            size = tbl.size()
            tbl.delete()
            return size

        assert len(port_ids) == 11
        assert _tbl_task() == 11

    def test_async_multiple_managers_queue_process_multiple_ports(self):
        tbl = Table({"a": int, "b": float, "c": str})
        tbl2 = Table({"a": int, "b": float, "c": str})
        port_ids = [0]
        port_data = [{"a": 0, "b": 0, "c": "0"}]

        for i in range(10):
            port_id = tbl.make_port()
            port_id2 = tbl2.make_port()
            assert port_id == port_id2
            port_ids.append(port_id)
            port_data.append({"a": port_id, "b": port_id * 1.5, "c": str(port_id)})

        manager = PerspectiveManager()
        manager2 = PerspectiveManager()
        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        manager.set_loop_callback(TestAsync.loop.add_callback)
        manager2.set_loop_callback(TestAsync.loop.add_callback)

        @syncify
        def _task():
            random.shuffle(port_ids)
            for port_id in port_ids:
                idx = port_id if port_id < len(port_ids) else len(port_ids) - 1
                tbl.update([port_data[idx]], port_id=port_id)
                tbl2.update([port_data[idx]], port_id=port_id)
            return (tbl.size(), tbl2.size())

        assert _task() == (11, 11)

    def test_async_multiple_managers_mixed_queue_process_multiple_ports(self):
        sentinel = {"async": 0, "sync": 0}

        def _counter(key, f, *args, **kwargs):
            sentinel[key] += 1
            return f(*args, **kwargs)

        sync_process = partial(_counter, "sync")
        async_process = partial(TestAsync.loop.add_timeout, 1, _counter, "async")
        tbl = Table({"a": int, "b": float, "c": str})
        tbl2 = Table({"a": int, "b": float, "c": str})
        port_ids = [0]
        port_data = [{"a": 0, "b": 0, "c": "0"}]

        for i in range(10):
            port_id = tbl.make_port()
            port_id2 = tbl2.make_port()
            assert port_id == port_id2
            port_ids.append(port_id)
            port_data.append({"a": port_id, "b": port_id * 1.5, "c": str(port_id)})

        manager = PerspectiveManager()
        manager2 = PerspectiveManager()
        manager.host_table("tbl", tbl)
        manager2.host_table("tbl2", tbl2)

        # manager uses tornado, manager2 is synchronous
        manager.set_loop_callback(async_process)
        manager2.set_loop_callback(sync_process)
        random.shuffle(port_ids)

        @syncify
        def _task():
            for port_id in port_ids:
                idx = port_id if port_id < len(port_ids) else len(port_ids) - 1
                tbl.update([port_data[idx]], port_id=port_id)

        _task()
        for port_id in port_ids:
            idx = port_id if port_id < len(port_ids) else len(port_ids) - 1
            tbl2.update([port_data[idx]], port_id=port_id)

        @syncify
        def _get_size():
            size = tbl.size()
            tbl.delete()
            return size

        assert _get_size() == 11
        assert tbl2.size() == 11
        assert sentinel["async"] == 2
        assert sentinel["sync"] == 12
        tbl2.delete()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .client import PerspectiveClient

try:
    from .aiohttp import PerspectiveAIOHTTPClient, websocket as aiohttp_websocket
except ImportError:
    ...

try:
    from .tornado import PerspectiveTornadoClient, websocket as tornado_websocket
except ImportError:
    ...



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from .libpsp import *
from .core import *
from .core._version import __version__
from .client import *
from .handlers import *
from .nbextension import _jupyter_nbextension_paths
from .widget import *



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import numpy as np
import pandas as pd


def _parse_datetime_index(index):
    """Given an instance of `pandas.DatetimeIndex`, parse its `freq` and
    return a `numpy.dtype` that corresponds to the unit it should be parsed in.

    Because `pandas.DataFrame`s cannot store datetimes in anything other than
    `datetime64[ns]`, we need to examine the `DatetimeIndex` itself to
    understand what unit it needs to be parsed as.

    Args:
        index (pandas.DatetimeIndex)

    Returns:
        `numpy.dtype`: a datetime64 dtype with the correct units depending on
            `index.freq`.
    """
    if index.freq is None:
        return np.dtype("datetime64[ns]")

    freq = str(index.freq).lower()
    new_type = None

    if any(s in freq for s in ["businessday", "day"]) or freq == "sm" or freq == "sms":
        # days
        new_type = "D"
    elif freq == "w" or "week" in freq:
        # weeks
        new_type = "W"
    elif any(s in freq for s in ["month", "quarter"]):
        # months
        new_type = "M"
    elif "year" in freq or freq == "a":
        new_type = "Y"
    else:
        # default to datetime
        new_type = "ns"

    return np.dtype("datetime64[{0}]".format(new_type))


def deconstruct_pandas(data, kwargs=None):
    """Given a dataframe, flatten it by resetting the index and memoizing the
    pivots that were applied.

    Args:
        data (pandas.dataframe): a Pandas DataFrame to parse

    Returns:
        (pandas.DataFrame, dict): a Pandas DataFrame and a dictionary containing
            optional members `columns`, `group_by`, and `split_by`.
    """
    kwargs = kwargs or {}
    kwargs = {"columns": [], "group_by": [], "split_by": []}

    # Decompose Period index to timestamps
    if isinstance(data.index, pd.PeriodIndex):
        data.index = data.index.to_timestamp()

    # convert categories to str
    if isinstance(data, pd.DataFrame):
        if hasattr(pd, "CategoricalDtype"):
            for k, v in data.dtypes.items():
                if isinstance(v, pd.CategoricalDtype):
                    data[k] = data[k].astype(str)

    if isinstance(data, pd.DataFrame) and isinstance(data.columns, pd.MultiIndex) and isinstance(data.index, pd.MultiIndex):
        # Row and col pivots
        kwargs["group_by"].extend([str(c) for c in data.index.names])

        # Two strategies
        if None in data.columns.names:
            # In this case, we need to extract the column names from the row
            # e.g. pt = pd.pivot_table(df, values = ['Discount','Sales'], index=['Country','Region'], columns=["State","Quantity"])
            # Table will be
            #                       Discount             Sales
            #         State       Alabama Alaska ...      Alabama Alaska ...
            #         Quantity    150 350 ...             300 500
            # Country Region
            #  US     Region 0    ...
            #  US     Region 1
            #
            # We need to transform this to:
            # group_by = ['Country', 'Region']
            # split_by = ['State', 'Quantity']
            # columns = ['Discount', 'Sales']
            existent = kwargs["group_by"] + data.columns.names
            for c in data.columns.names:
                if c is not None:
                    kwargs["split_by"].append(c)
                    data = data.stack()
            data = pd.DataFrame(data).reset_index()

            for new_column in data.columns:
                if new_column not in existent:
                    kwargs["columns"].append(new_column)
        else:
            # In this case, we have no need as the values is just a single entry
            # e.g. pt = pd.pivot_table(df, values = 'Discount', index=['Country','Region'], columns = ['Category', 'Segment'])
            for _ in kwargs["group_by"]:
                # unstack group by
                data = data.unstack()
            data = pd.DataFrame(data)

        # this rather weird loop is to map existing None columns into
        # levels, e.g. in the `else` block above, to reconstruct
        # the "Discount" name. IDK if this is stored or if the name is
        # lots, so we'll just call it 'index', 'index-1', ...
        i = 0
        new_names = list(data.index.names)
        for j, val in enumerate(data.index.names):
            if val is None:
                new_names[j] = "index" if i == 0 else "index-{}".format(i)
                i += 1
                # kwargs['group_by'].append(str(new_names[j]))
            else:
                if str(val) not in kwargs["group_by"]:
                    kwargs["split_by"].append(str(val))

        # Finally, remap any values columns to have column name 'value'
        data.index.names = new_names
        data = data.reset_index()  # copy
        data.columns = [str(c) if c in ["index"] + kwargs["group_by"] + kwargs["split_by"] + kwargs["columns"] else "value" for c in data.columns]
        kwargs["columns"].extend(["value" for c in data.columns if c not in ["index"] + kwargs["group_by"] + kwargs["split_by"] + kwargs["columns"]])
    elif isinstance(data, pd.DataFrame) and isinstance(data.columns, pd.MultiIndex):
        # Col pivots
        if data.index.name:
            kwargs["group_by"].append(str(data.index.name))
            push_group_by = False
        else:
            push_group_by = True

        data = pd.DataFrame(data.unstack())

        i = 0
        new_names = list(data.index.names)
        for j, val in enumerate(data.index.names):
            if val is None:
                new_names[j] = "index" if i == 0 else "index-{}".format(i)
                i += 1
                if push_group_by:
                    kwargs["group_by"].append(str(new_names[j]))
            else:
                if str(val) not in kwargs["group_by"]:
                    kwargs["split_by"].append(str(val))

        data.index.names = new_names
        data.columns = [str(c) if c in ["index"] + kwargs["group_by"] + kwargs["split_by"] else "value" for c in data.columns]
        kwargs["columns"].extend(["value" for c in data.columns if c not in ["index"] + kwargs["group_by"] + kwargs["split_by"]])

    elif isinstance(data, pd.DataFrame) and isinstance(data.index, pd.MultiIndex):
        # Group by
        kwargs["group_by"].extend(list(data.index.names))
        data = data.reset_index()  # copy

    if isinstance(data, pd.DataFrame):
        # flat df
        if "index" not in [str(c).lower() for c in data.columns]:
            data = data.reset_index(col_fill="index")

        if not kwargs["columns"]:
            # might already be set in row+col pivot df
            kwargs["columns"].extend([str(c) for c in data.columns])
            data.columns = kwargs["columns"]

    if isinstance(data, pd.Series):
        # Series
        flattened = data.reset_index()  # copy

        if isinstance(data, pd.Series):
            # preserve name from series
            flattened.name = data.name

            # make sure all columns are strings
            flattened.columns = [str(c) for c in flattened.columns]

        data = flattened

    return data, kwargs



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from pytest import raises
from perspective.core import PerspectiveError
from perspective.core import Plugin
import perspective.viewer.validate as validate


class TestValidate:

    def test_validate_plugin_valid_instance(self):
        assert validate.validate_plugin(Plugin.XBAR) == "X Bar"

    def test_validate_plugin_valid_instance_datagrid(self):
        assert validate.validate_plugin(Plugin.GRID) == "Datagrid"

    def test_validate_plugin_valid_string(self):
        assert validate.validate_plugin("X Bar") == "X Bar"

    def test_validate_plugin_invalid_string(self):
        with raises(PerspectiveError):
            validate.validate_plugin("invalid")

    def test_validate_plugin_invalid_string_hypergrid(self):
        with raises(PerspectiveError):
            validate.validate_plugin("hypergrid")

    def test_validate_filter_valid(self):
        filters = [["a", ">", 1], ["b", "==", "abc"]]
        assert validate.validate_filter(filters) == filters

    def test_validate_filter_invalid(self):
        with raises(PerspectiveError):
            filters = [["a", ">"], ["b", "invalid" "abc"]]
            validate.validate_filter(filters)

    def test_validate_filter_is_null(self):
        filters = [["a", "is null"]]
        assert validate.validate_filter(filters) == filters

    def test_validate_filter_is_not_null(self):
        filters = [["a", "is not null"]]
        assert validate.validate_filter(filters) == filters

    def test_validate_expressions(self):
        computed = ["// expression1 \n 'Hello'"]
        assert validate.validate_expressions(["// expression1 \n 'Hello'"]) == computed

    def test_validate_expressions_invalid(self):
        with raises(PerspectiveError):
            assert validate.validate_expressions({})



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import numpy as np
import pandas as pd
from perspective import PerspectiveViewer, Table


class TestViewer:

    def test_viewer_get_table(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        assert viewer.table == table

    # loading

    def test_viewer_load_table(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        assert viewer.columns == ["a"]

    def test_viewer_load_named_table(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table, name="data_1")
        assert viewer.columns == ["a"]
        assert viewer.table_name == "data_1"
        assert viewer.table == table

    def test_viewer_load_data(self):
        viewer = PerspectiveViewer()
        viewer.load({"a": [1, 2, 3]})
        assert viewer.columns == ["a"]

    def test_viewer_load_named_data(self):
        viewer = PerspectiveViewer()
        viewer.load({"a": [1, 2, 3]}, name="data_1")
        assert viewer.columns == ["a"]
        assert viewer.table_name == "data_1"

    def test_viewer_load_schema(self):
        viewer = PerspectiveViewer()
        viewer.load({
            "a": str,
            "b": int,
            "c": bool,
            "d": str
        })
        for col in viewer.columns:
            assert col in ["a", "b", "c", "d"]

    def test_viewer_load_table_with_options(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        # options should be disregarded when loading Table
        viewer.load(table, limit=1)
        assert viewer.columns == ["a"]
        assert viewer.table.size() == 3

    def test_viewer_load_data_with_options(self):
        viewer = PerspectiveViewer()
        # options should be forwarded to the Table constructor
        viewer.load({"a": [1, 2, 3]}, limit=1)
        assert viewer.columns == ["a"]
        assert viewer.table.size() == 1

    def test_viewer_load_clears_state(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer(theme="Pro Dark", group_by=["a"])
        viewer.load(table)
        assert viewer.group_by == ["a"]
        viewer.load({"b": [1, 2, 3]})
        assert viewer.group_by == []
        assert viewer.theme == "Pro Dark"  # should not break UI
    def test_viewer_load_np(self):
        table = Table({"a": np.arange(1, 100)})
        viewer = PerspectiveViewer()
        viewer.load(table)
        assert viewer.columns == ["a"]

    def test_viewer_load_np_data(self):
        viewer = PerspectiveViewer()
        viewer.load({"a": np.arange(1, 100)})
        assert viewer.columns == ["a"]
        assert viewer.table.size() == 99

    def test_viewer_load_df(self):
        table = Table(pd.DataFrame({"a": np.arange(1, 100)}))
        viewer = PerspectiveViewer()
        viewer.load(table)
        for col in viewer.columns:
            assert col in ["index", "a"]
        assert viewer.table.size() == 99

    def test_viewer_load_df_data(self):
        viewer = PerspectiveViewer()
        viewer.load(pd.DataFrame({"a": np.arange(1, 100)}))
        for col in viewer.columns:
            assert col in ["index", "a"]

    # update

    def test_viewer_update_dict(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.update({"a": [4, 5, 6]})
        assert table.size() == 6
        assert viewer.table.size() == 6
        assert viewer.table.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6]
        }

    def test_viewer_update_list(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.update([{"a": 4}, {"a": 5}, {"a": 6}])
        assert table.size() == 6
        assert viewer.table.size() == 6
        assert viewer.table.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6]
        }

    def test_viewer_update_df(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.update(pd.DataFrame({"a": [4, 5, 6]}))
        assert table.size() == 6
        assert viewer.table.size() == 6
        assert viewer.table.view().to_dict() == {
            "a": [1, 2, 3, 4, 5, 6]
        }

    def test_viewer_update_dict_partial(self):
        table = Table({"a": [1, 2, 3], "b": [5, 6, 7]}, index="a")
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.update({"a": [1, 2, 3], "b": [8, 9, 10]})
        assert table.size() == 3
        assert viewer.table.size() == 3
        assert viewer.table.view().to_dict() == {
            "a": [1, 2, 3],
            "b": [8, 9, 10]
        }

    # clear

    def test_viewer_clear(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.clear()
        assert viewer.table.size() == 0
        assert viewer.table.schema() == {
            "a": int
        }

    # replace

    def test_viewer_replace(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer()
        viewer.load(table)
        viewer.replace({"a": [4, 5, 6]})
        assert viewer.table.size() == 3
        assert viewer.table.schema() == {
            "a": int
        }
        assert viewer.table.view().to_dict() == {
            "a": [4, 5, 6]
        }

    # reset

    def test_viewer_reset(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer(plugin="X Bar", filter=[["a", "==", 2]])
        viewer.load(table)
        assert viewer.filter == [["a", "==", 2]]
        viewer.reset()
        assert viewer.plugin == "Datagrid"
        assert viewer.filter == []

    # delete

    def test_viewer_delete(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer(plugin="X Bar", filter=[["a", "==", 2]])
        viewer.load(table)
        assert viewer.filter == [["a", "==", 2]]
        viewer.delete()
        assert viewer.table_name is None
        assert viewer.table is None

    def test_viewer_delete_without_table(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer(plugin="X Bar", filter=[["a", "==", 2]])
        viewer.load(table)
        assert viewer.filter == [["a", "==", 2]]
        viewer.delete(delete_table=False)
        assert viewer.table_name is not None
        assert viewer.table is not None
        assert viewer.filter == []

    def test_save_restore(self):
        table = Table({"a": [1, 2, 3]})
        viewer = PerspectiveViewer(
            plugin="X Bar",
            filter=[["a", "==", 2]],
            expressions=['"a" * 2']
        )
        viewer.load(table)

        # Save config
        config = viewer.save()
        assert viewer.filter == [["a", "==", 2]]
        assert config["filter"] == [["a", "==", 2]]
        assert viewer.plugin == "X Bar"
        assert config["plugin"] == "X Bar"
        assert config["expressions"] == ['"a" * 2']

        # reset configuration
        viewer.reset()
        assert viewer.plugin == "Datagrid"
        assert viewer.filter == []
        assert viewer.expressions == []

        # restore configuration
        viewer.restore(**config)
        assert viewer.filter == [["a", "==", 2]]
        assert viewer.plugin == "X Bar"
        assert viewer.expressions == ['"a" * 2']

    def test_save_restore_plugin_config(self):
        viewer = PerspectiveViewer(plugin="Datagrid", plugin_config={"columns": {"a": {"fixed": 4}}})
        config = viewer.save()

        assert config["plugin_config"] == {
            "columns": {
                "a": {
                    "fixed": 4
                }
            }
        }

        viewer.reset()
        assert viewer.plugin_config == {}

        viewer.restore(**config)
        assert viewer.plugin_config == config["plugin_config"]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from aiohttp import web, WSMsgType, WebSocketError

from .common import PerspectiveHandlerBase


class PerspectiveAIOHTTPHandler(PerspectiveHandlerBase):
    """PerspectiveAIOHTTPHandler is a drop-in implementation of Perspective.

    Use it inside AIOHTTP routing to create a server-side Perspective that is
    ready to receive websocket messages from the front-end `perspective-viewer`.

    The Perspective client and server will automatically keep the Websocket
    alive without timing out.

    Examples:
        >>> manager = PerspectiveManager()
        >>> async def websocket_handler(request):
        ...    handler = PerspectiveAIOHTTPHandler(manager=manager, request=request)
        ...    await handler.run()

        >>> app = web.Application()
        >>> app.router.add_get("/websocket", websocket_handler)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._request = kwargs.get("request")

    async def run(self) -> None:
        try:
            self._ws = web.WebSocketResponse()
            await self._ws.prepare(self._request)

            async for msg in self._ws:
                if msg.type == WSMsgType.TEXT:
                    await self.on_message(msg.data)
                if msg.type == WSMsgType.BINARY:
                    await self.on_message(msg.data)
        finally:
            self.on_close()

    async def write_message(self, message: str, binary: bool = False) -> None:
        try:
            if binary:
                await self._ws.send_bytes(message)
            else:
                await self._ws.send_str(message)
        except WebSocketError:
            # Ignore error
            ...

    # Use common docstring
    __init__.__doc__ = PerspectiveHandlerBase.__init__.__doc__



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import pandas
from functools import partial, wraps
from random import random

from .view_config import ViewConfig
from ._data_formatter import to_format, _parse_format_options
from ._constants import COLUMN_SEPARATOR_STRING
from ._utils import _str_to_pythontype
from ._callback_cache import _PerspectiveCallBackCache
from ._date_validator import _PerspectiveDateValidator
from .libbinding import (
    make_view_unit,
    make_view_zero,
    make_view_one,
    make_view_two,
    to_arrow_unit,
    to_arrow_zero,
    to_arrow_one,
    to_arrow_two,
    to_csv_unit,
    to_csv_zero,
    to_csv_one,
    to_csv_two,
    get_row_delta_unit,
    get_row_delta_zero,
    get_row_delta_one,
    get_row_delta_two,
    scalar_to_py,
)


class View(object):
    """A :class:`~perspective.View` object represents a specific transform
    (pivot, filter, sort, etc) configuration on an underlying
    :class:`~perspective.Table`.  :class:`~perspective.View` objects
    cannot be directly instantiated - they must be derived from an existing
    :class:`~perspective.Table` via the :func:`~perspective.Table.view()`
    method.

    :class:`~perspective.View` instances receive all updates from the
    :class:`~perspective.Table` from which they are derived, and can be
    serialized (via ``to_*`` methods) or trigger a callback when it is updated.
    :class:`~perspective.View` objects will remain in memory and actively
    process updates until :obj:`~perspective.View.delete()` method is called.
    """

    def __init__(self, Table, **kwargs):
        self._name = "py_" + str(random())
        self._table = Table
        self._config = ViewConfig(**kwargs)
        self._sides = self.sides()

        date_validator = _PerspectiveDateValidator()

        self._is_unit_context = (
            self._table._index == ""
            and self._sides == 0
            and len(self._config.get_group_by()) == 0
            and len(self._config.get_split_by()) == 0
            and len(self._config.get_filter()) == 0
            and len(self._config.get_sort()) == 0
            and len(self._config.get_expressions()) == 0
        )

        if self._is_unit_context:
            self._view = make_view_unit(
                self._table._table,
                self._name,
                COLUMN_SEPARATOR_STRING,
                self._config,
                date_validator,
            )
        elif self._sides == 0:
            self._view = make_view_zero(
                self._table._table,
                self._name,
                COLUMN_SEPARATOR_STRING,
                self._config,
                date_validator,
            )
        elif self._sides == 1:
            self._view = make_view_one(
                self._table._table,
                self._name,
                COLUMN_SEPARATOR_STRING,
                self._config,
                date_validator,
            )
        else:
            self._view = make_view_two(
                self._table._table,
                self._name,
                COLUMN_SEPARATOR_STRING,
                self._config,
                date_validator,
            )

        self._column_only = self._view.is_column_only()
        self._update_callbacks = self._table._update_callbacks
        self._delete_callbacks = _PerspectiveCallBackCache()
        self._client_id = None

    def get_config(self):
        """Returns a copy of the immutable configuration ``kwargs`` from which
        this :class:`~perspective.View` was instantiated.

        Returns:
            :obj:`dict`: ``kwargs`` supplied to the
                :func:`perspective.Table.view()` method.
        """
        return self._config.get_config()

    def sides(self):
        """An integer representing the # of hierarchial axis on this
        :class:`~perspective.View`.

        0 - Neither ``group_by`` nor ``split_by`` properties are set.

        1 - ``group_by`` is set.

        2 - ``split_by`` is set (and also maybe ``group_by``).

        Returns:
            :obj:`int`: 0 <= N <= 2
        """
        if len(self._config.get_group_by()) > 0 or len(self._config.get_split_by()) > 0:
            if len(self._config.get_split_by()) > 0:
                return 2
            else:
                return 1
        else:
            return 0

    def get_min_max(self, colname):
        """Calculates the [min, max] of the leaf nodes of a column `colname`.

        Args:
            colname (:obj:`str`): The name of the column to calcualte range for.

        Returns:
            :obj:`list` of 2 elements, the `min` and `max` of the
        """
        return list(map(lambda x: scalar_to_py(x, False, False), self._view.get_min_max(colname)))

    def num_rows(self):
        """The number of aggregated rows in the :class:`~perspective.View`.

        This count includes the total aggregate rows for all ``group_by``
        depth levels, and can also be affected by any applied ``filter``.

        Returns:
            :obj:`int`: Number of rows.
        """
        return self._view.num_rows()

    def num_columns(self):
        """The number of aggregated columns in the :class:`~perspective.View`.
        This is affected by the ``split_by`` that are applied to the
        :class:`~perspective.View`.

        Returns:
            :obj:`int`: Number of columns.
        """
        return self._view.num_columns()

    def dimensions(self):
        """The View and Table aggregated dimensions in the :class:`~perspective.View`.

        Returns:
            :obj:`dict`: A dictionary with 4 fields indicating the dimensions of
                both the :class:`~perspective.View` and underlying
                :class:`~perspective.Table` objects: "num_view_rows",
                "num_view_columns", "num_table_rows", "num_table_columns"
        """
        return {
            "num_view_rows": self._view.num_rows(),
            "num_view_columns": self._view.num_columns(),
            "num_table_rows": self._table.num_rows(),
            "num_table_columns": self._table.num_columns(),
        }

    def get_row_expanded(self, idx):
        """Returns whether row at `idx` is expanded or collapsed.

        Returns:
            :obj:`bool`: Is this row expanded?
        """
        return self._view.get_row_expanded(idx)

    def expand(self, idx):
        """Expands the row at 'idx', i.e. displaying its leaf rows.

        Args:
            idx (:obj:`int`): Row index to expand.
        """
        return self._view.expand(idx, len(self._config.get_group_by()))

    def collapse(self, idx):
        """Collapses the row at 'idx', i.e. hiding its leaf rows.

        Args:
            idx (:obj:`int`): Row index to collapse.
        """
        return self._view.collapse(idx)

    def set_depth(self, depth):
        """Sets the expansion depth of the pivot tree.

        Args:
            depth (:obj:`int`): Depth to collapse all nodes to, which
                may be no greater then the length of the ``group_by``
                property.
        """
        return self._view.set_depth(depth, len(self._config.get_group_by()))

    def column_paths(self):
        """Returns the names of the columns as they show in the
        :class:`~perspective.View`, i.e. the hierarchial columns when
        ``split_by`` is applied.

        Returns:
            :obj:`list` of :obj`str`: Aggregated column names.
        """
        paths = self._view.column_paths()
        string_paths = []

        for path in paths:
            string_paths.append(COLUMN_SEPARATOR_STRING.join([p.to_string(False) for p in path]))

        return string_paths

    def schema(self, as_string=False):
        """The schema of this :class:`~perspective.View`, which is a key-value
        map that contains the column names and their Python data types.

        If the columns are aggregated, their aggregated types will be shown
        returned instead.

        Keyword Args:
            as_string (:obj:`bool`): returns data types as string
                representations, if ``True``.

        Returns:
            :obj:`dict`: A map of :obj:`str` column name to :obj:`str` or
                :obj:`type`, depending on the value of ``as_string`` kwarg.
        """
        if as_string:
            return {item[0]: item[1] for item in self._view.schema().items()}

        return {item[0]: _str_to_pythontype(item[1]) for item in self._view.schema().items()}

    def expression_schema(self, as_string=False):
        """Returns the expression schema of this :class:`~perspective.View`,
        which is a key-value map that contains the expressions and their
        Python data types.

        If the columns are aggregated, their aggregated types will be returned
        instead.

        Keyword Args:
            as_string (:obj:`bool`): returns data types as string
                representations, if ``True``.

        Returns:
            :obj:`dict`: A map of :obj:`str` column name to :obj:`str` or
                :obj:`type`, depending on the value of ``as_string`` kwarg.
        """
        if as_string:
            return {item[0]: item[1] for item in self._view.expression_schema().items()}

        return {item[0]: _str_to_pythontype(item[1]) for item in self._view.expression_schema().items()}

    def on_update(self, callback, mode=None):
        """Add a callback to be fired when :func:`perspective.Table.update()` is
        called on the parent :class:`~perspective.Table`.

        Multiple callbacks can be set through calling ``on_update`` multiple
        times, and will be called in the order they are set. Callback must be a
        callable function that takes exactly 1 or 2 parameters, depending on
        whether `on_update` is called with `mode="row"`. The first parameter is
        always `port_id`, an :obj:`int` that indicates which input port the
        update comes from. A `RuntimeError` will be thrown if the callback
        has mis-configured parameters.

        Args:
            callback (:obj:`callable`): a callable function reference that will
                be called when :func:`perspective.Table.update()` is called.
            mode (:obj:`str`): if set to "row", the callback will be passed
                an Arrow-serialized dataset of the rows that were updated.
                Defaults to "none".

        Examples:
            >>> def updater(port_id):
            ...     print("Update fired on port", port_id)
            >>> def updater_with_delta(port_id, delta):
            ...     print("Update on port", port_id, "delta len:", len(delta)))
            >>> view.on_update(updater)
            >>> view.on_update(updater, mode="row")
            >>> table.update({"a": [1]})'
            >>> Update fired on port 0
            >>> Update on port 0 delta len: 64
        """
        self._table._state_manager.call_process(self._table._table.get_id())
        mode = mode or "none"

        if not callable(callback):
            raise ValueError("Invalid callback - must be a callable function")

        if mode not in ["none", "row"]:
            raise ValueError('Invalid update mode {} - valid on_update modes are "none" or "row"'.format(mode))

        if mode == "row":
            if not self._view._get_deltas_enabled():
                self._view._set_deltas_enabled(True)

        wrapped_callback = partial(self._wrapped_on_update_callback, mode=mode, callback=callback)

        self._update_callbacks.add_callback(
            {
                "name": self._name,
                "orig_callback": callback,
                "callback": wrapped_callback,
            }
        )

    def remove_update(self, callback):
        """Given a callback function, remove it from the list of callbacks.

        Args:
            callback (:obj:`func`): a function reference that will be removed.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view2 = table.view()
            >>> def callback():
            ...     print("called!")
            >>> view.on_update(callback)
            >>> view2.on_update(callback)
            >>> table.update(new_data)
            called!
            >>> view2.remove_update(callback)
            >>> table.update(new_data) # callback removed and will not fire
        """
        self._table._state_manager.call_process(self._table._table.get_id())
        if not callable(callback):
            return ValueError("remove_update callback should be a callable function!")
        self._update_callbacks.remove_callbacks(lambda cb: cb["orig_callback"] == callback)

    def on_delete(self, callback):
        """Set a callback to be run when the :func:`perspective.View.delete()`
        method is called on this :class:`~perspective.View`.

        Args:
            callback (:obj:`callable`): A callback to run after
                :func:`perspective.View.delete()` method has been called.

        Examples:
            >>> def deleter():
            >>>     print("Delete called!")
            >>> view.on_delete(deleter)
            >>> view.delete()
            >>> Delete called!
        """
        if not callable(callback):
            return ValueError("on_delete callback must be a callable function!")
        self._delete_callbacks.add_callback(callback)

    def delete(self):
        """Delete the :class:`~perspective.View` and clean up all associated
        callbacks.

        This method must be called to clean up callbacks used by the
        :class:`~perspective.View`, as well as allow for deletion of the
        underlying :class:`~perspective.Table`.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view.delete()
        """
        self._table._state_manager.remove_process(self._table._table.get_id())
        self._table._views.pop(self._table._views.index(self._name))
        # remove the callbacks associated with this view
        self._update_callbacks.remove_callbacks(lambda cb: cb["name"] == self._name)
        [cb() for cb in self._delete_callbacks]

    def remove_delete(self, callback):
        """Remove the delete callback associated with this
        :class:`~perspective.View`.

        Args:
            callback (:obj:`callable`): A reference to a callable function that
                will be removed from delete callbacks.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view2 = table.view()
            >>> def callback():
            ...     print("called!")
            >>> view.on_delete(callback)
            >>> view2.on_delete(callback)
            >>> view.delete()
            called!
            >>> view2.remove_delete(callback)
            >>> view2.delete() # callback removed and will not fire
        """
        if not callable(callback):
            return ValueError("remove_delete callback should be a callable function!")
        self._delete_callbacks.remove_callbacks(lambda cb: cb == callback)

    def to_arrow(self, **kwargs):
        options = _parse_format_options(self, kwargs)
        if self._is_unit_context:
            return to_arrow_unit(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        elif self._sides == 0:
            return to_arrow_zero(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        elif self._sides == 1:
            return to_arrow_one(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        else:
            return to_arrow_two(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )

    def to_records(self, **kwargs):
        """Serialize the :class:`~perspective.View`'s dataset into a :obj:`list`
        of :obj:`dict` containing each row.

        By default, the entire dataset is returned, though this can be windowed
        via ``kwargs``.  When ``group_by`` are applied, a ``__ROW_PATH__``
        column name will be generated in addition to the applied ``columns``.
        When ``split_by`` are applied, column names will be qualified
        with their column group name.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            id (:obj:`bool`): Whether to return a logical row ID for each
                row (Defaults to ``False``).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`list` of :obj:`dict`: A list of :obj:`dict`, where each dict
                represents a row of the current state of the
                :class:`~perspective.View`.
        """
        return to_format(kwargs, self, "records")

    def to_dict(self, **options):
        """Serialize the :class:`~perspective.View`'s dataset into a :obj:`dict`
        of :obj:`str` keys and :obj:`list` values.  Each key is a column name,
        and the associated value is the column's data packed into a :obj:`list`.
        If the :class:`~perspective.View` is aggregated, the aggregated dataset
        will be returned.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            id (:obj:`bool`): Whether to return a logical row ID for each
                row (Defaults to ``False``).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`dict`: A dictionary with string keys and list values, where
                key = column name and value = column values.
        """
        return to_format(options, self, "dict")

    def to_numpy(self, **options):
        """Serialize the view's dataset into a :obj:`dict` of :obj:`str` keys
        and :class:`numpy.array` values.  Each key is a column name, and the
        associated value is the column's data packed into a numpy array.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            id (:obj:`bool`): Whether to return a logical row ID for each
                row (Defaults to ``False``).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`dict` of :class:`numpy.array`: A dictionary with string keys
                and numpy array values, where key = column name and
                value = column values.
        """
        return to_format(options, self, "numpy")

    def to_df(self, **options):
        """Serialize the view's dataset into a pandas dataframe.

        If the view is aggregated, the aggregated dataset will be returned.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            id (:obj:`bool`): Whether to return a logical row ID for each
                row (Defaults to ``False``).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).


        Returns:
            :class:`pandas.DataFrame`: A DataFrame serialization of the current
                state of this :class:`~perspective.View`.
        """
        cols = self.to_numpy(**options)
        return pandas.DataFrame(cols)

    def to_csv(self, **kwargs):
        """Serialize the :class:`~perspective.View`'s dataset into a CSV string.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            id (:obj:`bool`): Whether to return a logical row ID for each
                row (Defaults to ``False``).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to False).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to False).
            date_format (:obj:`str`): How ``datetime`` objects should be
                formatted in the CSV.

        Returns:
            :obj:`str`: A CSV-formatted string containing the serialized data.
        """

        options = _parse_format_options(self, kwargs)
        if self._is_unit_context:
            return to_csv_unit(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        elif self._sides == 0:
            return to_csv_zero(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        elif self._sides == 1:
            return to_csv_one(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )
        else:
            return to_csv_two(
                self._view,
                options["start_row"],
                options["end_row"],
                options["start_col"],
                options["end_col"],
            )

    @wraps(to_records)
    def to_json(self, **options):
        return self.to_records(**options)

    @wraps(to_dict)
    def to_columns(self, **options):
        return self.to_dict(**options)

    def _get_row_delta(self):
        if self._is_unit_context:
            return get_row_delta_unit(self._view)
        elif self._sides == 0:
            return get_row_delta_zero(self._view)
        elif self._sides == 1:
            return get_row_delta_one(self._view)
        else:
            return get_row_delta_two(self._view)

    def _num_hidden_cols(self):
        """Returns the number of columns that are sorted but not shown."""
        hidden = 0
        columns = self._config.get_columns()
        for sort in self._config.get_sort():
            if sort[0] not in columns:
                hidden += 1
        return hidden

    def _wrapped_on_update_callback(self, **kwargs):
        """Provide the user-defined callback function with additional metadata
        from the view.
        """
        mode = kwargs["mode"]
        port_id = kwargs["port_id"]
        cache = kwargs["cache"]
        callback = kwargs["callback"]

        if cache.get(port_id) is None:
            cache[port_id] = {}

        if mode == "row":
            if cache[port_id].get("row_delta") is None:
                cache["row_delta"] = self._get_row_delta()
            callback(port_id, cache["row_delta"])
        else:
            callback(port_id)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import date, datetime

import perspective
from pytest import mark


class TestTableInfer(object):

    def test_table_limit_wraparound_does_not_respect_partial(self):
        t = perspective.Table({'a':float, 'b':float}, limit=3)
        t.update([{'a':10}, {'b':1}, {'a':20}, {'a':None,'b':2}])
        df = t.view().to_df()

        t2 = perspective.Table({'a':float, 'b':float}, limit=3)
        t2.update([{'a':10}, {'b':1}, {'a':20}, {'b':2}])
        df2 = t2.view().to_df()

        assert df.to_dict() == df2.to_dict()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from perspective.table import Table
import psutil
import os


class TestDelete(object):

    # delete

    def test_table_delete(self):
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss
        for x in range(10000):
            data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
            tbl = Table(data)
            tbl.delete()
        
        mem2 = process.memory_info().rss

        # assert 1 < (max2 / max) < 1.01
        assert (mem2 - mem) < 2000000

    def test_table_delete_with_view(self, sentinel):
        data = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        tbl = Table(data)

        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss
        for x in range(10000):
            view = tbl.view()
            view.delete()
        
        tbl.delete()
        mem2 = process.memory_info().rss
        assert (mem2 - mem) < 2000000



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import datetime

import numpy as np

DATE_DTYPES = [
    np.dtype("datetime64[D]"),
    np.dtype("datetime64[W]"),
    np.dtype("datetime64[M]"),
    np.dtype("datetime64[Y]"),
]


def make_null_mask(array):
    """Given a numpy array, return a numpy array of int64s containing the
    indices of `array` where the value is either invalid or null.

    Invalid values are:
        - None
        - numpy.nat
        - numpy.nan

    Args:
        array (:obj:`numpy.array`)
    """
    mask = []

    is_object_or_string_dtype = np.issubdtype(array.dtype, np.str_) or np.issubdtype(array.dtype, np.object_)

    is_datetime_dtype = np.issubdtype(array.dtype, np.datetime64) or np.issubdtype(array.dtype, np.timedelta64)

    for i, item in enumerate(array):
        invalid = item is None

        if not is_object_or_string_dtype:
            if is_datetime_dtype:
                invalid = invalid or str(item) == "NaT"
            else:
                invalid = invalid or np.isnan(item)

        if invalid:
            mask.append(i)

    return mask


def deconstruct_numpy(array, mask=None):
    """Given a numpy array, parse it and return the data as well as a numpy
    array of null indices.

    Args:
        array (:obj:`numpy.array`)

    Keyword Args:
        mask (:obj:`numpy.array`)

    Returns:
        (:obj:`dict`): `array` is the original array, and `mask` is an array of
            booleans where `True` represents a nan/None value.
    """
    if mask is None:
        mask = make_null_mask(array)

    if array.dtype == bool or array.dtype == "?":
        # bool => byte
        array = array.astype("b", copy=False)
    elif np.issubdtype(array.dtype, np.datetime64):
        # treat days/weeks/months/years as datetime objects - avoid idiosyncracy
        # with days of month, etc.
        if array.dtype in DATE_DTYPES:
            array = array.astype(datetime)

        # cast datetimes to millisecond timestamps
        # because datetime64("nat") is a double, cast to float64 here - C++
        # handles the rest
        if array.dtype == np.dtype("datetime64[us]"):
            array = array.astype(np.float64, copy=False) / 1000
        elif array.dtype == np.dtype("datetime64[ns]"):
            array = array.astype(np.float64, copy=False) / 1000000
        elif array.dtype == np.dtype("datetime64[ms]"):
            array = array.astype(np.float64, copy=False)
        elif array.dtype == np.dtype("datetime64[s]"):
            array = array.astype(np.float64, copy=False) * 1000
        elif array.dtype == np.dtype("datetime64[m]"):
            array = array.astype(np.float64, copy=False) * 60000
        elif array.dtype == np.dtype("datetime64[h]"):
            array = array.astype(np.float64, copy=False) * 3600000
    elif np.issubdtype(array.dtype, np.timedelta64):
        array = array.astype(np.float64, copy=False)

    return {"array": array, "mask": mask}



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import aiohttp
import asyncio

from .websocket import (
    PerspectiveWebsocketClient,
    PerspectiveWebsocketConnection,
    Periodic,
)


class AIOHTTPPeriodic(Periodic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._running = True

    async def _run(self):
        while self._running:
            await self._callback()
            await asyncio.sleep(self._interval)

    async def start(self):
        return asyncio.create_task(self._run())

    async def stop(self):
        self._running = False


class PerspectiveAIOHTTPWebsocketConnection(PerspectiveWebsocketConnection):
    def __init__(self, session=None):
        self._ws = None
        self._session = session or aiohttp.ClientSession()
        self._run = True

    async def _receive_messages(self):
        async for msg in self._ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                self._on_message(msg.data)
            elif msg.type == aiohttp.WSMsgType.BINARY:
                self._on_message(msg.data)
            elif msg.type == aiohttp.WSMsgType.CLOSE:
                return

    async def connect(self, url, on_message, max_message_size) -> None:
        self._ws_cm = self._session.ws_connect(url)
        self._ws = await self._ws_cm.__aenter__()
        self._on_message = on_message
        self._task = asyncio.create_task(self._receive_messages())

    def periodic(self, callback, interval) -> Periodic:
        return AIOHTTPPeriodic(callback=callback, interval=interval)

    async def write(self, message, binary=False):
        if binary:
            return await self._ws.send_bytes(message)
        else:
            return await self._ws.send_str(message)

    async def close(self):
        try:
            self._task.cancel()
            await self._task
        except asyncio.CancelledError:
            ...
        await self._ws.close()


class PerspectiveAIOHTTPClient(PerspectiveWebsocketClient):
    def __init__(self, session=None):
        """Create a `PerspectiveAIOHTTPClient` that interfaces with a Perspective server over a Websocket"""
        super(PerspectiveAIOHTTPClient, self).__init__(PerspectiveAIOHTTPWebsocketConnection(session=session))


async def websocket(url, session=None):
    """Create a new websocket client at the given `url`.

    Args:
        session (:obj:`aiohttp.ClientSession`): An optional aiohtttp session
    """
    client = PerspectiveAIOHTTPClient(session=session)
    await client.connect(url)
    return client



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from starlette.testclient import TestClient
from starlette.websockets import WebSocketDisconnect

from queue import Empty

from .websocket import (
    PerspectiveWebsocketClient,
    PerspectiveWebsocketConnection,
    Periodic,
)


class _StarletteTestPeriodic(Periodic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Dont do anything as this should only ever be used in tests

    async def start(self):
        # Dont do anything as this should only ever be used in tests
        ...

    async def stop(self):
        # Dont do anything as this should only ever be used in tests
        ...


class _PerspectiveStarletteWebsocketConnection(PerspectiveWebsocketConnection):
    def __init__(self, client: TestClient):
        """A Starlette Websocket client.

        NOTE: For use in tests only!

        Args:
            client (TestClient): starlette TestClient instance
        """
        self._client = client
        self._ws = None
        self._on_message = None
        self._send_cache = None

    async def connect(self, url, on_message, max_message_size) -> None:
        self._ws = self._client.websocket_connect(url).__enter__()
        self._on_message = on_message

    def periodic(self, callback, interval) -> Periodic:
        return _StarletteTestPeriodic(callback=callback, interval=interval)

    def _on_message_internal(self):
        # This is a hacky workaround for the test client's fundamentally synchronous nature
        attempt = 0
        try_count = 5
        while attempt < try_count:
            try:
                while True:
                    # increase to avoid races
                    message = self._ws._send_queue.get(timeout=0.01)

                    if isinstance(message, BaseException):
                        raise message

                    self._ws._raise_on_close(message)

                    if "text" in message:
                        self._on_message(message["text"])
                    if "bytes" in message:
                        self._on_message(message["bytes"])

            except Empty:
                attempt += 1

    async def write(self, message, binary=False):
        # read back message
        self._on_message_internal()

        if binary:
            self._ws.send_bytes(message)
        else:
            self._ws.send_text(message)

        # read back message
        self._on_message_internal()

    async def close(self):
        try:
            self._ws.__exit__()
        except WebSocketDisconnect:
            return


class _PerspectiveStarletteTestClient(PerspectiveWebsocketClient):
    def __init__(self, test_client: TestClient):
        """Create a `PerspectiveStarletteTestClient` that interfaces with a Perspective server over a Websocket"""
        super(_PerspectiveStarletteTestClient, self).__init__(_PerspectiveStarletteWebsocketConnection(test_client))


async def websocket(test_client: TestClient, url: str):
    """Create a new websocket client at the given `url` using the thread current
    tornado loop."""
    client = _PerspectiveStarletteTestClient(test_client)
    await client.connect(url)
    return client



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

# NOTE: see test_tornado_lock for important notes
import asyncio
import pytest
import random

from datetime import datetime

from fastapi import FastAPI, WebSocket
from fastapi.testclient import TestClient

from perspective import (
    Table,
    PerspectiveManager,
    PerspectiveStarletteHandler,
)

from perspective.client.starlette_test import websocket


data = {
    "a": [i for i in range(10)],
    "b": [i * 1.5 for i in range(10)],
    "c": [str(i) for i in range(10)],
    "d": [datetime(2020, 3, i, i, 30, 45) for i in range(1, 11)],
}

MANAGER = PerspectiveManager()


async def websocket_handler(websocket: WebSocket):
    handler = PerspectiveStarletteHandler(
        manager=MANAGER, websocket=websocket, chunk_size=500
    )
    await handler.run()


APPLICATION = FastAPI()
APPLICATION.add_api_websocket_route("/websocket", websocket_handler)

CLIENT = TestClient(APPLICATION)


class TestPerspectiveStarletteHandlerChunked(object):
    def setup_method(self):
        """Flush manager state before each test method execution."""
        MANAGER._tables = {}
        MANAGER._views = {}

    async def websocket_client(self):
        """Connect and initialize a websocket client connection to the
        Perspective starlette server.
        """
        return await websocket(CLIENT, "/websocket")

    @pytest.mark.asyncio
    async def test_aiohttp_handler_create_view_to_arrow_chunked(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        view = await table.view()
        output = await view.to_arrow()
        expected = await table.schema()

        assert Table(output).schema(as_string=True) == expected
        await client.terminate()

    @pytest.mark.asyncio
    async def test_starlette_handler_create_view_to_arrow_chunked(self):
        table_name = str(random.random())
        _table = Table(data)
        MANAGER.host_table(table_name, _table)

        client = await self.websocket_client()
        table = client.open_table(table_name)
        views = await asyncio.gather(*[table.view() for _ in range(5)])
        outputs = await asyncio.gather(*[view.to_arrow() for view in views])
        expected = await table.schema()

        for output in outputs:
            assert Table(output).schema(as_string=True) == expected

        await client.terminate()



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import numpy
from calendar import timegm
from datetime import date, datetime
from dateutil.parser import parse
from pytz import UTC
from pandas import Period
from re import search
from time import mktime
from .libbinding import t_dtype


def _normalize_timestamp(obj):
    """Convert a timestamp in seconds to milliseconds.

    If the input overflows, it is treated as milliseconds - otherwise it is
    treated as seconds and converted.
    """
    try:
        datetime.fromtimestamp(obj)
        return int(obj * 1000)
    except (ValueError, OverflowError, OSError):
        return int(obj)


class _PerspectiveDateValidator(object):
    """Validate and parse dates using the `dateutil` package."""

    def parse(self, datestring):
        """Return a datetime.datetime object containing the parsed date, or
        None if the date is invalid.

        If a ISO date string with a timezone is provided, there is no guarantee
        that timezones will be properly handled by the parser. Perspective
        stores and serializes times in UTC as a milliseconds
        since epoch timestamp. For more definitive timezone support, use
        `datetime.datetime` objects or `pandas.Timestamp` objects with the
        `timezone` property set.

        Args:
            datestring (:obj:`str`): the datestring to parse

        Returns:
            (:class:`datetime.date`/`datetime.datetime`/`None`): if parse is
                successful.
        """
        try:
            return parse(datestring)
        except (ValueError, OverflowError):
            return None

    def to_date_components(self, obj):
        """Return a dictionary of string keys and integer values for `year`,
        `month` (from 0 - 11), and `day`.

        This method converts both datetime.date and numpy.datetime64 objects
        that contain datetime.date.
        """
        if obj is None:
            return obj

        if isinstance(obj, (int, float)):
            obj = datetime.fromtimestamp(_normalize_timestamp(obj) / 1000)

        if isinstance(obj, numpy.datetime64):
            if str(obj) == "NaT":
                return None
            obj = obj.astype(datetime)

            if isinstance(obj, int):
                obj = datetime.fromtimestamp(obj / 1000000000)

        # Perspective stores month in `t_date` as an integer [0-11],
        # while Python stores month as [1-12], so decrement the Python value
        # before Perspective attempts to use it.
        return {"year": obj.year, "month": obj.month - 1, "day": obj.day}

    def to_timestamp(self, obj):
        """Returns an integer corresponding to the number of milliseconds since
        epoch in the local timezone.

        If the `datetime.datetime` object has a `timezone` property set, this
        method will convert the object into UTC before returning a timestamp.
        """
        if obj is None:
            return obj

        if obj.__class__.__name__ == "date":
            # handle updating datetime with date object
            obj = datetime(obj.year, obj.month, obj.day)

        if isinstance(obj, Period):
            # extract the start of the Period
            obj = obj.to_timestamp()

        # time.mktime converts local time, calendar.timegm uses UTC
        converter = mktime

        # timetuple returns local time, utctimetuple returns UTC
        to_timetuple = "timetuple"

        if hasattr(obj, "tzinfo") and obj.tzinfo is not None:
            # Convert all timezone-aware datetimes into UTC. If a datetime with
            # a valid `tzinfo` object has been passed in, this will read
            # `tzinfo` and convert the datetime into UTC. If the datetime has no
            # `tzinfo` object, it is assumed to be in local time, and will not
            # be converted into UTC.
            obj = obj.astimezone(UTC)
            converter = timegm
            to_timetuple = "utctimetuple"

        if isinstance(obj, numpy.datetime64):
            if str(obj) == "NaT":
                return None

            # astype(datetime) returns an int or a long (in python 2)
            obj = obj.astype(datetime)

            if isinstance(obj, date) and not isinstance(obj, datetime):
                # Convert numpy "datetime64[D/W/M/Y]", as they are converted
                # into `datetime.date`.
                return int((converter(getattr(obj, to_timetuple)())) * 1000)

            if isinstance(obj, int):
                return round(obj / 1000000)

        if isinstance(obj, (int, float, numpy.integer, numpy.float64, numpy.float32)):
            return _normalize_timestamp(obj)

        timetuple = getattr(obj, to_timetuple)()

        is_datetime_min = timetuple.tm_year == 1 and timetuple.tm_mon == 1 and timetuple.tm_mday == 1 and timetuple.tm_hour == 0 and timetuple.tm_min == 0 and timetuple.tm_sec == 0

        if is_datetime_min:
            # Return beginning of epoch when datetime is datetime.min
            return 0

        if timetuple.tm_year < 1900:
            # Use calendar.timegm to do conversion between timetuple and
            # seconds timestamp
            converter = timegm

        # At this point, aware datetime objects are in UTC, and naive datetime
        # objects have not been modified. When the timestamp is serialized
        # using `to_format`, it will be in *local time* - Pybind will
        # automatically localize any conversion to `datetime.datetime`
        # from C++ to Python.
        seconds_timestamp = converter(timetuple) + obj.microsecond / 1000000.0
        ms_timestamp = int(seconds_timestamp * 1000)
        return ms_timestamp

    def format(self, datestring):
        """Return either t_dtype.DTYPE_DATE or t_dtype.DTYPE_TIME depending on
        the format of the parsed date.

        If the parsed date is invalid, return t_dtype.DTYPE_STR to prevent
        further attempts at conversion.  Attempt to use heuristics about dates
        to minimize false positives, i.e. do not parse dates without separators.

        Args:
            datestring (:obj:'str'): the datestring to parse.
        """
        if isinstance(datestring, (bytes, bytearray)):
            datestring = datestring.decode("utf-8")
        has_separators = bool(search(r"[/. -]", datestring))  # match commonly-used date separators
        # match commonly-used date separators

        dtype = t_dtype.DTYPE_STR

        if has_separators:
            try:
                parsed = parse(datestring)
                if (parsed.hour, parsed.minute, parsed.second, parsed.microsecond) == (
                    0,
                    0,
                    0,
                    0,
                ):
                    dtype = t_dtype.DTYPE_DATE
                else:
                    dtype = t_dtype.DTYPE_TIME
            except (ValueError, OverflowError, TypeError):
                # unparsable dates should be coerced to string
                dtype = t_dtype.DTYPE_STR

        return dtype



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "nbextension/static",
            "dest": "@finos/perspective-jupyterlab",
            "require": "@finos/perspective-jupyterlab/extension",
        }
    ]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import datetime

from ..core import ALL_FILTERS, Aggregate, Plugin, Sort
from ..core.exception import PerspectiveError


def validate_plugin(plugin):
    if isinstance(plugin, Plugin):
        return plugin.value
    elif isinstance(plugin, str):
        if plugin not in Plugin.options():
            raise PerspectiveError("Unrecognized `plugin`: {0}".format(plugin))
        return plugin
    else:
        raise PerspectiveError("Cannot parse `plugin` of type: {0}".format(str(type(plugin))))


def validate_columns(columns):
    if columns is None:
        return []
    elif isinstance(columns, str):
        columns = [columns]

    if isinstance(columns, list):
        return columns
    else:
        raise PerspectiveError("Cannot parse `columns` of type: %s", str(type(columns)))


def _validate_pivots(pivots):
    if pivots is None:
        return []
    elif isinstance(pivots, str):
        pivots = [pivots]

    if isinstance(pivots, list):
        return pivots
    else:
        raise PerspectiveError("Cannot parse pivots of type: %s", str(type(pivots)))


def validate_group_by(group_by):
    return _validate_pivots(group_by)


def validate_split_by(split_by):
    return _validate_pivots(split_by)


def validate_aggregates(aggregates):
    if aggregates is None:
        return {}
    elif isinstance(aggregates, dict):
        for k, v in aggregates.items():
            if isinstance(v, Aggregate):
                aggregates[k] = v.value
            elif isinstance(v, str):
                if v not in Aggregate.options():
                    raise PerspectiveError("Unrecognized aggregate: %s", v)
            elif isinstance(v, list):
                # Parse weighted mean aggregate in ["weighted mean", "COLUMN"]
                if len(v) == 2 and v[0] == "weighted mean":
                    continue
                raise PerspectiveError(
                    "Unrecognized aggregate in incorrect syntax for weighted mean: %s - Syntax should be: ['weighted mean', 'COLUMN']",
                    v,
                )
            else:
                raise PerspectiveError("Cannot parse aggregation of type %s", str(type(v)))
        return aggregates
    else:
        raise PerspectiveError("Cannot parse aggregates type: %s", str(type(aggregates)))


def validate_sort(sort):
    if sort is None:
        return []
    elif isinstance(sort, str):
        sort = [sort]

    if isinstance(sort, list):
        if len(sort) > 0 and not isinstance(sort[0], list):
            sort = [sort]
        ret = []
        for col, s in sort:
            if isinstance(s, Sort):
                s = s.value
            elif not isinstance(s, str) or s not in Sort.options():
                raise PerspectiveError("Unrecognized sort direction: %s", s)
            ret.append([col, s])
        return ret
    else:
        raise PerspectiveError("Cannot parse sort type: %s", str(type(sort)))


def validate_filter(filters):
    if filters is None:
        return []

    elif isinstance(filters, list) and len(filters) > 0 and not isinstance(filters[0], list):
        # wrap
        filters = [filters]

    if isinstance(filters, list):
        for f in filters:
            if not isinstance(f, list):
                raise PerspectiveError("`filter` kwarg must be a list!")

            for i, item in enumerate(f):
                if i == 1:
                    if item not in ALL_FILTERS:
                        raise PerspectiveError("Unrecognized filter operator: {}".format(item))
                    elif item not in ("is null", "is not null"):
                        if len(f) != 3:
                            raise PerspectiveError("Cannot parse filter - {} operator must have a comparison value.".format(item))
                else:
                    if item.__class__.__name__ == "date":
                        f[i] = item.strftime("%m/%d/%Y")
                    elif isinstance(item, datetime):
                        f[i] = item.strftime("%Y-%m-%d %H:%M:%S")
        return filters
    else:
        raise PerspectiveError("Cannot parse filter type: {}".format(str(type(filters))))


def validate_expressions(expressions):
    if expressions is None:
        return []

    if isinstance(expressions, str):
        # wrap in a list and return
        return [expressions]

    if isinstance(expressions, list):
        for expr in expressions:
            if not isinstance(expr, str):
                raise PerspectiveError("Cannot parse non-string expression: {}".format(str(type(expr))))
        return expressions
    else:
        raise PerspectiveError("Cannot parse expressions of type: {}".format(str(type(expressions))))


def validate_plugin_config(plugin_config):
    return plugin_config or {}


def validate_title(title):
    return title
